From 538989aa4b12a975d435bd1838935bedf9f465ca Mon Sep 17 00:00:00 2001
From: Alexandre Cassen <acassen@corp.free.fr>
Date: Fri, 28 Feb 2025 17:24:58 +0100
Subject: net/mlx5e: Support routed networks during IPsec accel MACs init

Remote IPsec tunnel endpoint may refer to a network segment that is
not directly connected to the host. In such a case, IPsec tunnel
endpoints are connected to a router and reachable via a routing path.
In IPsec packet offload mode, HW is initialized with the MAC address
of both IPsec tunnel endpoints.

Extend the current IPsec init MACs procedure to resolve nexhop for
routed networks. Direct neighbour lookup and probe is still used
for directly connected networks and as a fallback mechanism if fib
lookup fails.

Signed-off-by: Alexandre Cassen <acassen@corp.free.fr>
Signed-off-by: Leon Romanovsky <leonro@nvidia.com>
---
 .../ethernet/mellanox/mlx5/core/en_accel/ipsec.c   | 36 ++++++++++++++++++++++
 1 file changed, 36 insertions(+)

diff --git a/drivers/net/ethernet/mellanox/mlx5/core/en_accel/ipsec.c b/drivers/net/ethernet/mellanox/mlx5/core/en_accel/ipsec.c
index 4c46fc4f0cad26..1060a3998986a9 100644
--- a/drivers/net/ethernet/mellanox/mlx5/core/en_accel/ipsec.c
+++ b/drivers/net/ethernet/mellanox/mlx5/core/en_accel/ipsec.c
@@ -284,8 +284,10 @@ static void mlx5e_ipsec_init_macs(struct mlx5e_ipsec_sa_entry *sa_entry,
 	struct mlx5_core_dev *mdev = mlx5e_ipsec_sa2dev(sa_entry);
 	struct xfrm_state *x = sa_entry->x;
 	struct net_device *netdev;
+	struct flowi4 fl4 = {};
 	struct neighbour *n;
 	u8 addr[ETH_ALEN];
+	struct rtable *rt;
 	const void *pkey;
 	u8 *dst, *src;
 
@@ -294,17 +296,22 @@ static void mlx5e_ipsec_init_macs(struct mlx5e_ipsec_sa_entry *sa_entry,
 		return;
 
 	netdev = x->xso.real_dev;
+	fl4.flowi4_proto = x->sel.proto;
 
 	mlx5_query_mac_address(mdev, addr);
 	switch (attrs->dir) {
 	case XFRM_DEV_OFFLOAD_IN:
 		src = attrs->dmac;
 		dst = attrs->smac;
+		fl4.daddr = attrs->addrs.saddr.a4;
+		fl4.saddr = attrs->addrs.daddr.a4;
 		pkey = &attrs->addrs.saddr.a4;
 		break;
 	case XFRM_DEV_OFFLOAD_OUT:
 		src = attrs->smac;
 		dst = attrs->dmac;
+		fl4.daddr = attrs->addrs.daddr.a4;
+		fl4.saddr = attrs->addrs.saddr.a4;
 		pkey = &attrs->addrs.daddr.a4;
 		break;
 	default:
@@ -312,6 +319,33 @@ static void mlx5e_ipsec_init_macs(struct mlx5e_ipsec_sa_entry *sa_entry,
 	}
 
 	ether_addr_copy(src, addr);
+
+	/* Destination can refer to a routed network.
+	 * Do a fib lookup to resolve nexthop and get
+	 * its mac. Neighbour resolution is used as
+	 * fallback.
+	 */
+	rt = ip_route_output_key(dev_net(netdev), &fl4);
+	if (IS_ERR(rt))
+		goto neigh;
+
+	if (rt->rt_type != RTN_UNICAST) {
+		ip_rt_put(rt);
+		goto neigh;
+	}
+
+	n = dst_neigh_lookup(&rt->dst, &fl4.daddr);
+	if (!n) {
+		ip_rt_put(rt);
+		goto neigh;
+	}
+
+	neigh_ha_snapshot(addr, n, netdev);
+	ether_addr_copy(dst, addr);
+	ip_rt_put(rt);
+	goto neigh_release;
+
+neigh:
 	n = neigh_lookup(&arp_tbl, pkey, netdev);
 	if (!n) {
 		n = neigh_create(&arp_tbl, pkey, netdev);
@@ -323,6 +357,8 @@ static void mlx5e_ipsec_init_macs(struct mlx5e_ipsec_sa_entry *sa_entry,
 		neigh_ha_snapshot(addr, n, netdev);
 		ether_addr_copy(dst, addr);
 	}
+
+neigh_release:
 	neigh_release(n);
 }
 
-- 
cgit 1.2.3-korg
